"""A creation workflow that adds existing Super PoP switches to the coreDB."""

import ipaddress

from orchestrator import workflow
from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle
from orchestrator.workflow import StepList, done, init, step
from orchestrator.workflows.steps import resync, set_status, store_process_subscription

from gso.products import ProductType
from gso.products.product_types import super_pop_switch
from gso.products.product_types.super_pop_switch import SuperPopSwitchInactive
from gso.services import subscriptions
from gso.services.crm import get_customer_by_name
from gso.services.subscriptions import get_site_by_name
from gso.utils.helpers import generate_fqdn
from gso.utils.shared_choices import PortNumber, Vendor


@step("Create subscription")
def create_subscription(customer: str) -> State:
    """Create a new subscription object."""
    customer_id = get_customer_by_name(customer)["id"]
    product_id = subscriptions.get_product_id_by_name(ProductType.SUPER_POP_SWITCH)
    subscription = SuperPopSwitchInactive.from_product_id(product_id, customer_id)

    return {
        "subscription": subscription,
        "subscription_id": subscription.subscription_id,
    }


def initial_input_form_generator() -> FormGenerator:
    """Generate a form that is filled in using information passed through the :term:`API` endpoint."""

    class ImportSuperPopSwitch(FormPage):
        class Config:
            title = "Import a Super PoP switch"

        customer: str
        super_pop_switch_site: str
        hostname: str
        super_pop_switch_ts_port: PortNumber
        super_pop_switch_mgmt_ipv4_address: ipaddress.IPv4Address

    user_input = yield ImportSuperPopSwitch

    return user_input.dict()


@step("Initialize subscription")
def initialize_subscription(
    subscription: SuperPopSwitchInactive,
    hostname: str,
    super_pop_switch_ts_port: PortNumber,
    super_pop_switch_site: str,
    super_pop_switch_mgmt_ipv4_address: ipaddress.IPv4Address | None = None,
) -> State:
    """Initialise the Super PoP switch subscription using input data."""
    subscription.super_pop_switch.super_pop_switch_ts_port = super_pop_switch_ts_port
    site_obj = get_site_by_name(super_pop_switch_site).site
    subscription.super_pop_switch.super_pop_switch_site = site_obj
    fqdn = generate_fqdn(hostname, site_obj.site_name, site_obj.site_country_code)
    subscription.super_pop_switch.super_pop_switch_fqdn = fqdn
    subscription.description = f"Super PoP switch {fqdn}"
    subscription.super_pop_switch.super_pop_switch_mgmt_ipv4_address = super_pop_switch_mgmt_ipv4_address
    subscription.super_pop_switch.vendor = Vendor.JUNIPER

    subscription = super_pop_switch.SuperPopSwitchProvisioning.from_other_lifecycle(
        subscription, SubscriptionLifecycle.PROVISIONING
    )

    return {"subscription": subscription}


@workflow(
    "Import Super PoP switch",
    initial_input_form=initial_input_form_generator,
    target=Target.CREATE,
)
def import_super_pop_switch() -> StepList:
    """Import a Super PoP switch without provisioning it."""
    return (
        init
        >> create_subscription
        >> store_process_subscription(Target.CREATE)
        >> initialize_subscription
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
